#ifndef __relorient5m_c /* Copyright Bill.Triggs@inrialpes.fr */
#define __relorient5m_c "$ Fri Jul 21 12:32:44 2000 $"
/** NO WARRANTY. Use & distribute freely. Fixes/feedback welcome. **/
/*-------------------------------------------------------------------
  int relorient5m(double qout[20][4], double Rtout[20][3][4], 
  %		  double dout[20*2*n], double y[n][3], double x[n][3], int n)

  Relative orientation of two calibrated cameras from n=5 unknown 3D
  points, using sparse multiresultant. The 5 input point pairs are in
  y[][] and x[][]. Each point should be normalized as a unit-norm
  direction vector. The output is the k<=20 consistent solutions for the
  relative orientation of the cameras and the point depths. In a frame
  where the calibrated 3x4 projection matrix P1 (giving x) has form
  (I|0), for solution s, the second camera (giving y) has form P2 = Rt =
  (R|t), where Rt is given by Rtout[s], R=R(q) is a 3x3 rotation matrix
  with quaternion q given by qout[s]. The corresponding point depths are
  in dout[2*n*k+p] for x[p] and dout[2*n*k+p+n] for y[p], where 0<=p<n
  is the point label. For each solution, t and the depths are scaled so
  that the closest point has unit depth, i.e. 3D distances are measured
  in units of the smallest point-camera distance.

  Singularities: The method works for coplanar points, but becomes
  ill-conditioned for small translations |t| << |depths|. In this case,
  the returned depths are always inaccurate and some of them may become
  negative, as a small perturbation of t would change its sign, and
  hence the signs of the depths. A heuristic fix for this case would be
  to multiply the negative depths by -1 or -S for S>1, but we do not do
  this.

  The underlying method also has singularities for certain rotation
  values, notably for rotations very close to zero. Ideally this would
  be fixed by another multiresultant reduction aimed at this
  case. Instead we work around the problem by applying a random rotation
  to y before calculating R, then undoing it later. The area of
  ill-conditioning is quite small, so this does not seem to cause random
  problems for non-zero rotations in practice.
  
  Number of solutions: There are 20 (10 twisted pairs) in theory, but
  many are often complex. The twisted solutions always have inconsistent
  signs for all their depths, so we reduce immediately to at most 10
  solutions, and more often 2-4 in practice.

  Method
  ------
  Related but not identical to that reported in [1].  Let x be the
  direction vector and dx the distance to a 3D point in camera frame 1,
  and similarly y, dy in image 2. Let the motion between the camera
  frames be R(q),t, where q is a quaternion parametrizing the 3x3
  inter-frame rotation R and t is the translation between the camera
  centres. Then

  %	dy * y = R * (dx * x) + t			(1)

  Also, the points will be visible iff dx>0 and dy>0.  The coplanarity
  constraint is

  %	[y,R*x,t] = 0 

  where [,,] is the triple product. By the usual quaternion rotation
  formula, R*x = q*x*C(q) where C(q) is the conjugate quaternion. Also
  the triple product is the scalar part of the triple quaternion
  product, so the coplanarity constraint can be written

  %	(y * q * x * C(q) * t)_0 = (y * q * x * qt)_0 = 0  

  where x,y,t are viewed as quaternions with vanishing scalar part, the
  product is quaternion product, (..)_0 is the scalar part, and qt is
  the quaternion C(q)*t. This constraint can also be written

  %	q' * B(x,y) * qt = 0 

  where we regard q and qt as column 4-vectors and B is the 4x4
  matrix

  %	B(x,y) = [ x*y' + y*x' - (y'*x)*eye(3), y^x; -(y^x)', -(y'*x) ]

  We have one of these constraints for each pair of corresponding
  points (x,y), and also the constraint

  %	(qt * q)_0 = qt[0]*q[0]-sum(qt[i]*q[i],i=1..3) = 0

  which says that R'*t is a pure vector (vanishing scalar part), i.e. qt
  and q are consistent.

  For N points this gives us N+1 bilinear polynomials in the 4+4=8
  unknowns qt and q, with known coefficients (depending on the observed
  points x,y). We use a sparse multivariable resultant to reduce the
  solution of this problem to an eigenvector problem as follows (see
  reference [1] for the general method). The system has 20 roots in 10
  twisted pairs (i.e. differing by a relative 180 degree rotation about
  the line between the camera centres), but some of these are often
  complex.

  I used Maple to apply Emeris' `far' routine [1] to the above N+1
  polynomials with variable ordering [qt[1],qt[2],q[1],q[2],q[3]] (for
  normalization we take qt[4]=1=q[4], where q[4] is the scalar component
  of q). This gives a 60x60 matrix with 10 rows for each input
  polynomial, the coefficients being linear polynomials in the hidden
  variable qt[3], with coefficients either +-1 or an entry of a B
  matrix. Only 20 of the columns involve qt[3] -- permute these to the
  right hand side of the matrix.

  The first 10 rows (qt*q constraint) have 0's and 1's as coefficients,
  so Gaussian elimination based on these leaves a 50x50 matrix, still
  with 20 columns containing qt[3]. This 50x50 matrix is built below
  (M0,M1), by stacking A0 (10x50) and A1 (10x20) matrices, one for each
  point pair, where A1 contains the coefficients of qt[3] and A0 the
  constant coefficients.

  Now, continue Gaussian elimination of the first 30 columns of M0,
  applying the decomposition to the remaining 20 columns of M0 and to
  M1. This gives 20x20 matrices R0,R1. For solutions of the original
  problem, we have

  %	det(qt[4]*R0 + qt[3]*R1) = 0,

  In fact the solutions (q,t) to the original system can be read off
  from the generalized eigenvectors of the matrix pencil (R0,R1), whose
  components are proportional to the 20 cubic monomials in q[i],i=1..4.
  Ideally, we would use a generalized eigenvector routine for this, but
  for the moment we just find R0^-1*R1 and use a standard eigenvector
  routine. (The result is in principle equivalent, but less stable if R0
  is ill-conditioned).  Given q and hence R, t is easy found from the
  coplanarity constraints.  The depths dx,dy can be recovered linearly
  from (1) above as usual.

  The method fails:
  1) When the first 30 columns of M0 are not linearly independent or when
  R0 is not full rank (it is not clear when these happen, but using
  a generalized eigenvector routine would get around some rank deficiency
  of R0).
  2) When qt[3]/qt[4] becomes infinite, and probably also when
  |(qt[3],qt[4])| << |(qt[1],qt[2])|. NB: qt = (q[4]*t-q^t,q'*t) where
  ^ is cross product and q is the scaled rotation axis. 
  3) When two or more solutions happen to have the same value of qt[3]/qt[4].
  4) If t is very small compared to the distance to the scene.

  The failure 4 is intrinsic. For t->0 the depths can not be found
  because _any_ depths are consistent with the given R.  Failures 1-3
  are reduced by applying a random rotation to y before calling the
  routine, then undoing it afterwards. Changes of variable ordering and
  hidden variable would also have a similar effect, but require another
  version of the same routine.

  References: 
  [1] Emiris, Ioannis Z., A General Solver Based on Sparse Resultants:
  Numerical Issues and Kinematic Applications, INRIA Research Report
  3110, January 1997, http://www.inria.fr/RRRT/RR-3110.html
-----------------------------------------------------------------------*/
#include <stdlib.h>
/*#include <stdio.h>*/
#include <math.h>
#include "geom23.h"

extern void 
  LAPACK(dgesvd)(), 
  LAPACK(dgeqrf)(),
  LAPACK(dgeevx)(), 
  LAPACK(dgetrf)(), 
  LAPACK(dgetrs)();

/* If q = (X,Y,Z,W) is quaternion of R, the multiresultant eigenvectors
 * are proportional to:
 *
 * { X^3,   X^2*Y, X^2*Z, X*Y^2, X*Y*Z, X*Z^2, Y^3,   Y^2*Z, Y*Z^2, Z^3, 
 *   X^2*W, X*Y*W, X*Z*W, Y^2*W, Y*Z*W, Z^2*W, X*W^2, Y*W^2, Z*W^2, W^3 }
 *
 * Here are the indices of the (q[X]..q[W])*(something) monomials
 * in the eigenvector, used to read off q from the eigenvector.  
 */
static const unsigned char 
qmon[4][4] = {{0,1,2,10},	/* q[X]^2*q */
	     {3,6,7,13},	/* q[Y]^2*q */
	     {5,8,9,15},	/* q[Z]^2*q */
	     {16,17,18,19}};	/* q[W]^2*q */

int
relorient5m(double qout[20][4], double Rtout[20][3][4], double dout[/*20*2*n*/], 
	    double y[/*n*/][3], double x[/*n*/][3], int n)
{
  long dn = 10*n, lwork = 20*dn, info;
  double *M0 = TMPALLOC(double,dn*(50+20+20));
  double *M1 = M0 + dn*50, *work = M1 + dn*20, qq[4],RR[3][3];
  int p,sol,nsol, randomize = 1;
  
  /* The below multiresultant formulation has an annoying singularity
   * for rotations very close to zero, so we move this singularity
   * elsewhere by applying a random rotation to the y points.
   */
  if (randomize) {
    for (p=X; p<=Z; p++) qq[p] = randn();
    qq[W] = 1.0;
    quat_to_rot(RR,qq,1);
  }
  
  /* Build up the 50x50 and 50x20 resultant matrices M0,M1. There are 10
   * rows for each point pair, each row representing a different
   * multiple of the pairs' coplanarity constraint polynomial
   * q'*B(x,y)*qt.  
   */
  for (p=0; p< dn*70; p++) M0[p] = 0.0;

  for (p=0; p<n; p++) {
    double *_A0 = M0 + 10*p-(dn+1), *_A1 = M1 + 10*p-(dn+1);
    double *yp = y[p], *xp = x[p], Ry[3], B[4][4];

    if (randomize) {
      m33v3(Ry,RR,yp);
      yp = Ry;
    }
    /* B = [ x*y' + y*x' - (y'*x)*eye(3), y^x; -(y^x)', -(y'*x) ] 
     */
    B[0][0] = yp[X]*xp[X]-yp[Y]*xp[Y]-yp[Z]*xp[Z];
    B[0][1] = yp[X]*xp[Y]+yp[Y]*xp[X];
    B[0][2] = yp[X]*xp[Z]+yp[Z]*xp[X];
    B[0][3] = yp[Y]*xp[Z]-yp[Z]*xp[Y];
    B[1][0] = yp[X]*xp[Y]+yp[Y]*xp[X];
    B[1][1] = -yp[X]*xp[X]+yp[Y]*xp[Y]-yp[Z]*xp[Z];
    B[1][2] = yp[Y]*xp[Z]+yp[Z]*xp[Y];
    B[1][3] = -yp[X]*xp[Z]+yp[Z]*xp[X];
    B[2][0] = yp[X]*xp[Z]+yp[Z]*xp[X];
    B[2][1] = yp[Y]*xp[Z]+yp[Z]*xp[Y];
    B[2][2] = -yp[X]*xp[X]-yp[Y]*xp[Y]+yp[Z]*xp[Z];
    B[2][3] = yp[X]*xp[Y]-yp[Y]*xp[X];
    B[3][0] = -yp[Y]*xp[Z]+yp[Z]*xp[Y];
    B[3][1] = yp[X]*xp[Z]-yp[Z]*xp[X];
    B[3][2] = -yp[X]*xp[Y]+yp[Y]*xp[X];
    B[3][3] = -yp[X]*xp[X]-yp[Y]*xp[Y]-yp[Z]*xp[Z];
    
#define A0(i,j) _A0[i+dn*(j)]
#define A1(i,j) _A1[i+dn*(j)]

    A0(1,24) = B[1][0];
    A0(1,25) = B[2][0];
    A0(1,26) = B[0][1];
    A0(1,27) = B[1][1]-B[0][0];
    A0(1,28) = B[2][1];
    A0(1,29) = B[3][0];
    A0(1,30) = B[3][1];
    A0(1,47) = B[0][3];
    A0(1,48) = B[1][3];
    A0(1,49) = B[2][3];
    A0(1,50) = B[3][3]+B[0][0];
    A0(2,16) = B[1][0];
    A0(2,17) = B[2][0];
    A0(2,20) = B[0][1];
    A0(2,22) = B[1][1]-B[0][0];
    A0(2,23) = B[2][1];
    A0(2,25) = B[3][0];
    A0(2,28) = B[3][1];
    A0(2,43) = B[0][3];
    A0(2,45) = B[1][3];
    A0(2,46) = B[2][3];
    A0(2,49) = B[3][3]+B[0][0];
    A0(3,3) = B[1][0];
    A0(3,4) = B[2][0];
    A0(3,10) = B[0][1];
    A0(3,13) = B[1][1]-B[0][0];
    A0(3,14) = B[2][1];
    A0(3,17) = B[3][0];
    A0(3,23) = B[3][1];
    A0(3,36) = B[0][3];
    A0(3,39) = B[1][3];
    A0(3,40) = B[2][3];
    A0(3,46) = B[3][3]+B[0][0];
    A0(4,15) = B[1][0];
    A0(4,16) = B[2][0];
    A0(4,19) = B[0][1];
    A0(4,21) = B[1][1]-B[0][0];
    A0(4,22) = B[2][1];
    A0(4,24) = B[3][0];
    A0(4,27) = B[3][1];
    A0(4,42) = B[0][3];
    A0(4,44) = B[1][3];
    A0(4,45) = B[2][3];
    A0(4,48) = B[3][3]+B[0][0];
    A0(5,2) = B[1][0];
    A0(5,3) = B[2][0];
    A0(5,9) = B[0][1];
    A0(5,12) = B[1][1]-B[0][0];
    A0(5,13) = B[2][1];
    A0(5,16) = B[3][0];
    A0(5,22) = B[3][1];
    A0(5,35) = B[0][3];
    A0(5,38) = B[1][3];
    A0(5,39) = B[2][3];
    A0(5,45) = B[3][3]+B[0][0];
    A0(6,1) = B[1][0];
    A0(6,2) = B[2][0];
    A0(6,8) = B[0][1];
    A0(6,11) = B[1][1]-B[0][0];
    A0(6,12) = B[2][1];
    A0(6,15) = B[3][0];
    A0(6,21) = B[3][1];
    A0(6,34) = B[0][3];
    A0(6,37) = B[1][3];
    A0(6,38) = B[2][3];
    A0(6,44) = B[3][3]+B[0][0];
    A0(7,18) = B[0][1];
    A0(7,19) = B[1][1]-B[0][0];
    A0(7,20) = B[2][1];
    A0(7,21) = -B[1][0];
    A0(7,22) = -B[2][0];
    A0(7,26) = B[3][1];
    A0(7,27) = -B[3][0];
    A0(7,41) = B[0][3];
    A0(7,42) = B[1][3];
    A0(7,43) = B[2][3];
    A0(7,47) = B[3][3]+B[0][0];
    A0(7,48) = B[1][0];
    A0(7,49) = B[2][0];
    A0(7,50) = B[3][0];
    A0(8,7) = B[0][1];
    A0(8,9) = B[1][1]-B[0][0];
    A0(8,10) = B[2][1];
    A0(8,12) = -B[1][0];
    A0(8,13) = -B[2][0];
    A0(8,20) = B[3][1];
    A0(8,22) = -B[3][0];
    A0(8,33) = B[0][3];
    A0(8,35) = B[1][3];
    A0(8,36) = B[2][3];
    A0(8,43) = B[3][3]+B[0][0];
    A0(8,45) = B[1][0];
    A0(8,46) = B[2][0];
    A0(8,49) = B[3][0];
    A0(9,6) = B[0][1];
    A0(9,8) = B[1][1]-B[0][0];
    A0(9,9) = B[2][1];
    A0(9,11) = -B[1][0];
    A0(9,12) = -B[2][0];
    A0(9,19) = B[3][1];
    A0(9,21) = -B[3][0];
    A0(9,32) = B[0][3];
    A0(9,34) = B[1][3];
    A0(9,35) = B[2][3];
    A0(9,42) = B[3][3]+B[0][0];
    A0(9,44) = B[1][0];
    A0(9,45) = B[2][0];
    A0(9,48) = B[3][0];
    A0(10,5) = B[0][1];
    A0(10,6) = B[1][1]-B[0][0];
    A0(10,7) = B[2][1];
    A0(10,8) = -B[1][0];
    A0(10,9) = -B[2][0];
    A0(10,18) = B[3][1];
    A0(10,19) = -B[3][0];
    A0(10,31) = B[0][3];
    A0(10,32) = B[1][3];
    A0(10,33) = B[2][3];
    A0(10,41) = B[3][3]+B[0][0];
    A0(10,42) = B[1][0];
    A0(10,43) = B[2][0];
    A0(10,47) = B[3][0];

    A1(1,17) = B[0][2];
    A1(1,18) = B[1][2];
    A1(1,19) = B[2][2]-B[0][0];
    A1(1,20) = B[3][2];
    A1(2,13) = B[0][2];
    A1(2,15) = B[1][2];
    A1(2,16) = B[2][2]-B[0][0];
    A1(2,19) = B[3][2];
    A1(3,6) = B[0][2];
    A1(3,9) = B[1][2];
    A1(3,10) = B[2][2]-B[0][0];
    A1(3,16) = B[3][2];
    A1(4,12) = B[0][2];
    A1(4,14) = B[1][2];
    A1(4,15) = B[2][2]-B[0][0];
    A1(4,18) = B[3][2];
    A1(5,5) = B[0][2];
    A1(5,8) = B[1][2];
    A1(5,9) = B[2][2]-B[0][0];
    A1(5,15) = B[3][2];
    A1(6,4) = B[0][2];
    A1(6,7) = B[1][2];
    A1(6,8) = B[2][2]-B[0][0];
    A1(6,14) = B[3][2];
    A1(7,11) = B[0][2];
    A1(7,12) = B[1][2];
    A1(7,13) = B[2][2]-B[0][0];
    A1(7,15) = -B[1][0];
    A1(7,16) = -B[2][0];
    A1(7,17) = B[3][2];
    A1(7,19) = -B[3][0];
    A1(8,3) = B[0][2];
    A1(8,5) = B[1][2];
    A1(8,6) = B[2][2]-B[0][0];
    A1(8,9) = -B[1][0];
    A1(8,10) = -B[2][0];
    A1(8,13) = B[3][2];
    A1(8,16) = -B[3][0];
    A1(9,2) = B[0][2];
    A1(9,4) = B[1][2];
    A1(9,5) = B[2][2]-B[0][0];
    A1(9,8) = -B[1][0];
    A1(9,9) = -B[2][0];
    A1(9,12) = B[3][2];
    A1(9,15) = -B[3][0];
    A1(10,1) = B[0][2];
    A1(10,2) = B[1][2];
    A1(10,3) = B[2][2]-B[0][0];
    A1(10,5) = -B[1][0];
    A1(10,6) = -B[2][0];
    A1(10,11) = B[3][2];
    A1(10,13) = -B[3][0];
#undef A0
#undef A1
  }
  /* Reduce the (10*n)x50 system M0 + lambda*[zeros(10*n,30),M1] to a
   * 20x20 system I + lambda*R1. We factor M0 by pivoted LU
   * decomposition, then solve the linear system M0 * X = M1 on the rows
   * chosen by pivoting, to get R ~= M0^-1 * M1.
   */
  {
    long kk = 20, nn = 50, ipiv[50];
    LAPACK(dgetrf)(&dn,&nn,M0,&dn,ipiv,&info);
    LAPACK(dgetrs)("N",&dn,&kk,M0,&dn,ipiv,M1,&dn,&info);
  }
  {
    /* Solve the 20x20 eigensystem R1*x=lambda*x. (We could also use
     * a generalized eigensolver here for better stability).
     */
    long nn = 20, ilo = 1, ihi = nn, inone=1,jpvt[20];
    double *B = M1+30, alphaim[20], alphare[20], ZZ[20*20], 
      R[3][3], t[3], dnone[1], scale[20],abnrm[1],rconde[20],rcondv[20];
    int j;

    LAPACK(dgeevx)("N","N","V","N",&nn,B,&dn,alphare,alphaim,dnone,&inone,ZZ,&nn,
		   &ilo,&ihi,scale,abnrm,rconde,rcondv,work,&lwork,jpvt,&info);

    /* The columns of ZZ are the multiresultant eigenvectors, whose
     * entries are proportional to monomials in q. For each real
     * eigenvector, find the largest set of components
     * (q[X]..q[W])*(something), read off q from thier ratios, normalize
     * it, and find the rotation R(q).  
     */
    for (sol=nsol=0; sol<20; sol++) {
      if (alphaim[sol]==0.0) {
	double *z = ZZ + 20*sol, q[4], smax=-1.0;
	int i,imax=-1,ndneg;
	for (i=0; i<4; i++) {
	  double s = (SQ(z[qmon[i][0]])+SQ(z[qmon[i][1]])
		      +SQ(z[qmon[i][2]])+SQ(z[qmon[i][3]]));
	  if (s>smax) smax = s, imax = i;
	}
	smax = sqrt(smax);
	for (i=0; i<4; i++) q[i] = z[qmon[imax][i]]/smax;
	if (randomize) quat_left_div(q,qq,q);
	quat_to_rot(R,q,0);

	/* Now that we have q and R, solve for t using the coplanarity
	 * constraints [y[p],R*x[p],t] = 0 from all n points, where [,,]
	 * is the triple product.
	 */
	{
	  double (*Rx)[3] = (double(*)[3])M0, 
	    *A = M0 + 3*n, *S = A + 3*n, *d1 = S + n,
	    dmin=1e100;
	  long mm = n, nn = 3, info;

	  for (p=0; p<n; p++) m33v3(Rx[p],R,x[p]);
	  for (p=0; p<n; p++) {
	    A[0*n+p] = y[p][Y]*Rx[p][Z]-y[p][Z]*Rx[p][Y];	
	    A[1*n+p] = y[p][Z]*Rx[p][X]-y[p][X]*Rx[p][Z];
	    A[2*n+p] = y[p][X]*Rx[p][Y]-y[p][Y]*Rx[p][X];
	  }
	  LAPACK(dgesvd)("N","O",&mm,&nn,A,&mm,S,dnone,&inone,dnone,&inone,
			 work,&lwork,&info);
	  for (i=0; i<3; i++) t[i] = A[2+n*i];
	  
	  /* Given R,t, substitute into d2[p] * y[p] = R * (d1[p] *
	   * x[p]) + t and solve in least squares to find the depths
	   * d1[p],d2[p] of x[p] and y[p] for points 1..n. We may have
	   * had the wrong sign for t, so flip the depths if more than
	   * half are negative.  Also, for small translations, |t|<<|d|
	   * with |t| close to or below the noise level, some of the
	   * pairs may recommend t's of opposite signs above, owing to
	   * noise. When the average t is used to find d1,d2, these
	   * pairs will end up with negative d1,d2 because of this.
	   * These correspond to mathematically valid point pairs that
	   * happen (owing to noise) to be predicted to be so distant
	   * that they are `beyond infinity'. As this solution is
	   * mathematically valid, any method of correcting it
	   * (e.g. multiplying the depths by -k, for some arbitrary
	   * k>=1) must be heuristic, so we don't do it here. (It only
	   * happens for |t| close to the noise level anyway). However,
	   * although the absolute sign of d1,d2 is unreliable in this
	   * way, their relative sign should be reliable provided that
	   * neither approaches 0 or infinity. We use this to suppress
	   * solutions with impossible signs.
	   */
	  ndneg = 0;
	  for (p=0; p<n; p++) {
	    double C[3][3];
	    for (j=X; j<=Z; j++) {
	      C[0][j] = y[p][j];
	      C[1][j] = -Rx[p][j];
	      C[2][j] = -t[j];
	    }
	    LAPACK(dgesvd)("N","O",&nn,&nn,(double*)C,&nn,S,
			   dnone,&inone,dnone,&inone,work,&lwork,&info);

	    if ((C[0][2]<0.0) != (C[1][2]<0.0))	/* impossible sign, skip solution */
	      goto next_soln;

	    d1[p] = C[1][2] / C[2][2];
	    d1[p+n] = C[0][2] / C[2][2];
	    ndneg += (d1[p]<0.0) + (d1[p+n]<0.0);
	    if (dmin>fabs(d1[p])) dmin = fabs(d1[p]);
	    if (dmin>fabs(d1[p+n])) dmin = fabs(d1[p+n]);
	  }
	  /* Scale depths and t to distance to nearest point 
	   */
	  for (i=X; i<=Z; i++) t[i] /= dmin;
	  for (p=0; p<2*n; p++) d1[p] /= dmin;
	  
	  if (ndneg > n) {
	    /* >50% depths negative, flip solution 
	     */
	    for (i=X; i<=Z; i++) t[i] = -t[i];
	    for (p=0; p<2*n; p++) d1[p] = -d1[p];
	    ndneg = 2*n-ndneg;
	  }
	  /* A feasible solution: write out the q, Rt and point depths.
	   */
	  {
	    if (qout) 
	      for (i=X; i<=W; i++) qout[nsol][i] = q[i];
	    if (Rtout) {
	      for (i=X; i<=Z; i++) {
		for (j=X; j<=Z; j++) Rtout[nsol][i][j] = R[i][j];
		Rtout[nsol][i][W] = t[i];
	      }
	    }
	    if (dout) 
	      for (p=0; p<2*n; p++) dout[2*n*nsol+p] = d1[p];
	    nsol++;
	  }
	}
      }
    next_soln:
      continue;
    }
  }
  TMPFREE(M0);
  return nsol;
}

#endif /*__relorient5m_c */
