#ifndef __test_relorient5_c /* Copyright Bill.Triggs@inrialpes.fr */
#define __test_relorient5_c "$ Fri Jul  7 20:25:08 2000 $"
/** NO WARRANTY. Use & distribute freely. Fixes/feedback welcome. **/
/*------------------------------------------------------------------
  test_relorient5() -- hack test routine for relorient5().  Generates a
  random scene and set of camera positions, runs the relative
  orientation routine on them, and prints the ground-truth residuals of
  the solutions found.
 -------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "relorient5.c"
#include "relorient5m.c"
#include "relorient_approx_rot.c"
#include "rot_to_quat.c"
#include "quat_left_div.c"
#include "quat_to_rot.c"
#include "randn.c"
#include "vecs_to_rot.c"
#include "percentile.c"

/*------------------------------------------------------------------------
  Print m x n matrix (for debugging).
 ------------------------------------------------------------------------*/
void
pmat(double A[/*(i,j)->i+j*dn*/], int m, int n, int dn) 
{
  int i,j;
  if (dn<=0) dn=m;
  for (i=0; i<m; i++) {
    for (j=0; j<n; j++) fprintf(stdout," % #5g",A[i+j*dn]);
    fputc('\n',stdout);
  }
  fputc('\n',stdout);
}

void
pmat_t(double A[/*(i,j)->i+j*dn*/], int m, int n, int dn) 
{
  int i,j;
  if (dn<=0) dn=m;
  for (j=0; j<n; j++) {
    for (i=0; i<m; i++) 
      fprintf(stdout," % #5g",A[i+j*dn]);
    fputc('\n',stdout);
  }
  fputc('\n',stdout);
}

/*------------------------------------------------------------------------
  Write m x n matrix in MATLAB format (for debugging).
 ------------------------------------------------------------------------*/
void
pmatlab(char* file, char* name, double A[], int m, int n, int dn) 
{
  FILE* f = file? fopen(file,"a") : stdout;
  int i,j;
  fprintf(f,"%s = [",name?name:"A");
  if (dn<=0) dn=m;
  for (i=0; i<m; i++) {
    for (j=0; j<n; j++) fprintf(f," % #15g",A[i+j*dn]);
    fputc('\n',f);
  }
  fputs("];\n",f);
  if (file) fclose(f);
}

/*------------------------------------------------------------------------
  Normalize a vector or matrix to a unit vector.
 ------------------------------------------------------------------------*/
double
vnormalize(double v[], int n) 
{
  int i;
  double s = 1e-100;
  for (i=0; i<n; i++) s += SQ(v[i]);
  s = sqrt(s);
  for (i=0; i<n; i++) v[i] /= s;
  return s;
}


/*------------------------------------------------------------------------
  Read input points x,y from a file.  Format: an integer n giving the
  number of point pairs, then n coordinate triples for x, then n
  coordinate triples for y.    
 ------------------------------------------------------------------------*/
int
read_xy_file(char fname[], double x[][3], double y[][3], int nmax)
{
  FILE* in = strcmp(fname,"-")? fopen(fname,"r") : stdin;
  int p,n=-1;
  if (!in) {
    fprintf(stderr,"Can't open file `%s'\n",fname);
    return -1;
  }
  if (fscanf(in,"%d",&n) != 1 || n<5 || n>nmax) {
    fprintf(stderr,"Bad number of input points: `%d'\n",n);
    goto done;
  }
  for (p=0; p<n; p++) {
    if (fscanf(in,"%lf %lf %lf",&x[p][X],&x[p][Y],&x[p][Z]) != 3) {
      fprintf(stderr,"Can't read point x[%d]\n",p);
      goto done;
    }
  }
  for (p=0; p<n; p++) {
    if (fscanf(in,"%lf %lf %lf",&y[p][X],&y[p][Y],&y[p][Z]) != 3) {
      fprintf(stderr,"Can't read point y[%d]\n",p);
      goto done;
    }
  }
 done:
  if (in && in!=stdin) fclose(in);
  return n;
}

/*------------------------------------------------------------------------
  Print the various elements of a possible solution: 
  (R|t) = calibrated 3x4 projection matrix of camera 2 in frame of camera 1
  q = quaternion of R
  d = depths of n points x, then n points y.  
 ------------------------------------------------------------------------*/
void
print_solution(FILE* out, double q[4], double Rt[3][4], double d[/*2*n*/], int n)
{
  int i;
  fprintf(out,"(R|t):\n");
  for (i=0; i<3; i++) {
    fprintf(out,"%12g\t %12g\t %12g\t %12g\n", Rt[i][X],Rt[i][Y],Rt[i][Z],Rt[i][W]);	
  }
  fprintf(out,"q:\n%12g\t %12g\t %12g\t %12g\n", q[X],q[Y],q[Z],q[W]);
  fprintf(out,"dx:\n");
  for (i=0; i<n; i++) fprintf(out,"%8g\t ",d[i]);
  fprintf(out,"\ndy:\n");	
  for (i=0; i<n; i++) fprintf(out,"%8g\t ",d[i+n]);
  fprintf(out,"\n");	      
}

/*------------------------------------------------------------------------
  Quick and dirty getopt() for systems without a real one.  You'll need
  to #define NEED_getopt and also the character with which command line
  options start ('-' on UNIX, '/' on DOS I believe).
  ------------------------------------------------------------------------*/
#ifdef NEED_getopt
extern char* strchr();
char *optarg = 0;
int optargc = 0;
#ifdef DOS
# define OPTCHAR0 '/'
#else
# define OPTCHAR0 '-'
#endif

int 
getopt(int argc, char* argv[], char* flags)
{
  int f;
  char* p;
  /* skip non-flag args */
  do 
    if (++optargc>=argc) return EOF;
  while (argv[optargc][0] != OPTCHAR0);
  
  if ((f = argv[optargc][1]) 
      && (p = strchr(flags,f))) {
    optarg = (optargc<argc-1 && p[1] == ':')? argv[++optargc] : 0;
  } 
  return f;
}
#endif /* NEED_getopt */


/*------------------------------------------------------------------------
  Main routine, either reads point coordinates from a file and prints
  answer, or generates random scenes and summarizes estimation errors.
 ------------------------------------------------------------------------*/
#define NMAX 10
int
main(int argc, char* argv[])
{
  int n = 5,ntrials = 1, verbose = 2,
    i,j,p,nsol,trial;
  double 
    noise = 0.0, 
    unflatness = 1.0, 
    dist = 4.0, 
    spread = 1e0, 
    rot = 20.0,
    max_t_zero = 1.0,
    min_t_nonzero = 0.0;
  double xx[NMAX][3],x[NMAX][3],y[NMAX][3],dd[NMAX][2],
    P1[3][4],P2[3][4],P21[3][4],
    R0[3][3],R2[3][3],t1[3],t2[3],q0[4],q2[4],t21[3],t22[3],
    qout[20][4],Rtout[20][3][4],dout[NMAX*2*20],dmin;
  char* xy_file_name = 0;
  static  char *usage = 
    "Usage: [-N noise(radians)] [-f unflatness] [-t ntrials]\n"
    "[-d camera-distance] [-s camera-separation] [-a camera-angle]\n"
    "[-z zero|nonzero] [-M max_t_zero] [-m min_t_nonzero]\n"
    "[-i xy_file]\n";
  extern char *optarg;
  extern int getopt();

  while ((i = getopt(argc,argv,"a:d:f:i:M:m:n:N:s:t:v:z:")) != EOF) {
    switch (i) {
    case 'a':	rot = atof(optarg);		break;
    case 'd':   dist = atof(optarg);		break;
    case 'f':	unflatness = atof(optarg);	break;
    case 'i':	xy_file_name = optarg;		break;
    case 'm':	min_t_nonzero = atof(optarg);	break;
    case 'M':	max_t_zero = atof(optarg);	break;
   /* case 'n':	n = atoi(optarg);		break; */
    case 'N':	noise = atof(optarg);		break;
    case 's':	spread = atof(optarg);		break;
    case 't':	ntrials = atoi(optarg);		break;
    case 'v':	verbose = atoi(optarg);		break;
    case 'z':
      /* Force only the zero translation, or only the nonzero
       * translation method to be called.  
       */
      if (optarg[0]=='z') {
	max_t_zero = min_t_nonzero = 1e100;
      }
      else if (optarg[0]=='n') {
	max_t_zero = min_t_nonzero = -1.0;
      }
      break;
    case '?': 
    default:
      n=-1;
      break;
    }
  }
  if (n<5) {
    fprintf(stderr,usage);
    exit(1);
  }
    
  if (xy_file_name) {
    /* Read input points x,y from a file and print the corresponding
     * solutions.  
     */
    if ((n = read_xy_file(xy_file_name,x,y,NMAX)) < 5) 
      exit(1);

    nsol = relorient5(qout,Rtout,dout,y,x,n,max_t_zero,min_t_nonzero);

    for (i=0; i<nsol; i++) {
      fprintf(stdout,"\n--- Solution %d ---\n",i);
      print_solution(stdout,qout[i],Rtout[i],dout+2*i*n,n);
    }
  }
  else {
    /* Generate ntrials random scenes and run the method on each one.
     */
    double esum_R=0.0, esum_t=0.0, esum_d=0.0;
    double *med_R = TMPALLOC(double,3*ntrials), 
      *med_t = med_R+ntrials, *med_d = med_t+ntrials;
    
    for (trial=0; trial<ntrials; trial++) {

      /* Generate random camera rotations and translation.
       */
      for (i=X; i<=Z; i++) {
	q0[i] = randn()*rot*M_PI/180.0/sqrt(3.0);
	q2[i] = randn()*rot*M_PI/180.0/sqrt(3.0);
	t1[i] = 0.0;
	t2[i] = randn()*spread/sqrt(3.0);
      }
      q0[W] = q2[W] = 1.0;
      t1[Z] += dist;
      t2[Z] += dist;
      quat_to_rot(R0,q0,1);
      quat_to_rot(R2,q2,1);
      m33v3(t21,R2,t1);
      m33v3(t22,R2,t2);
      for (j=X; j<=Z; j++) t21[j] = t22[j] - t21[j];
      /* vnormalize(t21,3);*/

      /* Generate (calibrated) 3x4 camera projection matrices.
       */
      for (i=X; i<=Z; i++) {
	for (j=X; j<=Z; j++) {
	  P1[i][j] = (i==j);
	  P2[i][j] = P21[i][j] = R2[i][j];
	}
	P1[i][W] = t1[i];
	P2[i][W] = t22[i];
	P21[i][W] = t21[i];
      }

      /* Generate random 3D points, project them, and add noise. Also
       * find the minimum point-camera distance dmin as this gives the
       * implicit distance scale for the output depths and translation.  
       */
      for (p=0,dmin=1e37; p<n; p++) {
	double x1[3],d;
	x1[X] = randn();
	x1[Y] = randn();
	x1[Z] = unflatness*randn();
	m33v3(xx[p],R0,x1); 
	p34v3(x[p],P1,xx[p]);
	p34v3(y[p],P2,xx[p]);
      
	d = dd[p][0] = vnormalize(x[p],3);
	if (dmin>d) dmin = d;
	d = dd[p][1] = vnormalize(y[p],3);
	if (dmin>d) dmin = d;
	for (i=X; i<=Z; i++) x[p][i] += noise*randn(), y[p][i] += noise*randn();
      }
      for (i=X; i<=Z; i++) P21[i][W] /= dmin;
      for (p=0; p<n; p++) dd[p][0] /= dmin, dd[p][1] /= dmin;
#if 0
      pmatlab("MM.m","xx",(double*)xx,3,n,3);
      pmatlab("MM.m","x",(double*)x,3,n,3);
      pmatlab("MM.m","y",(double*)y,3,n,3);
      pmatlab("MM.m","P1",(double*)P1,4,3,4);
      pmatlab("MM.m","P2",(double*)P2,4,3,4);
#elif 0
      /* print points in input format */
      fprintf(stdout,"%d\n\n",n);
      pmat_t((double*)x,3,n,3);
      pmat_t((double*)y,3,n,3);
      fprintf(stdout,"\n");
      pmat_t((double*)P21,4,3,4);
#endif

      /* Call the orientation routine.
       */
      nsol = relorient5(qout,Rtout,dout,y,x,n,max_t_zero,min_t_nonzero);

      /* Evaluate the solutions against the ground truth, and maybe
       * print some results.  
       */
      {
	double Rerror[21],terror[21],derror[21], besterr = 1e10;
	int besti = -1;

	for (i=0; i<nsol; i++) {
	  double Rerr=0.0,terr=0.0, derr=0.0;
	  double (*Rt1)[4] = Rtout[i];
	  for (j=X; j<=Z; j++) {
	    Rerr += SQ(Rt1[j][X]-P21[j][X])+SQ(Rt1[j][Y]-P21[j][Y])+SQ(Rt1[j][Z]-P21[j][Z]);
	    terr += SQ(Rt1[j][W]-P21[j][W]);
	  }
	  for (p=0; p<n; p++) {
	    derr += SQ(1.0/dout[2*n*i+p]-1.0/dd[p][0]) + SQ(1.0/dout[2*n*i+p+n]-1.0/dd[p][1]);
	  }
	  /* Rotation, translation, and depth errors. In degrees, % of
	   * distance to nearest point, and average % error.
	   */
	  Rerr = 180.0/M_PI*sqrt(Rerr);
	  terr = 1e2*sqrt(terr);
	  derr = 100.0*sqrt(derr)/(2*n);
	  Rerror[i] = Rerr;
	  terror[i] = terr;
	  derror[i] = derr;
	  if (Rerr<besterr) {
	    besterr = Rerr;
	    besti = i;
	  }
	  if (verbose>1) {
	    fprintf(stdout,"%d.%d: dR = %12g deg\tdt = %12g %%\tdd = %12g %%\n",
		    trial,i,Rerr,terr,derr);
	    if (verbose>4) {
	      pmat((double*)Rtout[i],4,3,4);
	      pmat((double*)P21,4,3,4);
	      pmat(dout+2*n*i,1,2*n,1);
	    }
	  }
	}
	esum_R += med_R[trial] = Rerror[besti];
	esum_t += med_t[trial] = terror[besti];
	esum_d += med_d[trial] = derror[besti];

	if (verbose==1) {
	  fprintf(stdout,"%d %12g\t %12g\t %12g\n",
		  trial,Rerror[besti],terror[besti],derror[besti]);
	}
	else if (verbose>1) {
	  fprintf(stdout,"\n");
	}
      }
    }
#if 0
      pmatlab("TMP.m","med_R",med_R,1,ntrials,1);
      pmatlab("TMP.m","med_t",med_t,1,ntrials,1);
      pmatlab("TMP.m","med_d",med_d,1,ntrials,1);
#endif    
    fprintf(stdout,"%12g\t %12g\t %12g\t %12g\t %12g\t %12g\n",
	    esum_R/ntrials, dpercentile(ntrials/2,med_R,ntrials),
	    esum_t/ntrials, dpercentile(ntrials/2,med_t,ntrials),
	    esum_d/ntrials, dpercentile(ntrials/2,med_d,ntrials));
  }
  return 0;
}

#endif /*__test_relorient5_c */
